'use strict';

const {createHash} = require('crypto');
const {inspect} = require('util');
const {StringDecoder} = require('string_decoder');

const inspectWithKind = require('inspect-with-kind');
const isPlainObj = require('is-plain-obj');
const parseContentType = require('content-type').parse;
const Parse5SaxParser = require('parse5-sax-parser');

const TMP_HEADER_NAME = 'nodejs-temporary-inserted-header-name';
const PARSER_EVENTS = new Set(['startTag', 'endTag', 'comment', 'text', 'doctype']);
const CONTENT_LENGTH_ERROR = 'According to RFC7230, Content-Length header must be a non-negative integer https://tools.ietf.org/html/rfc7230#section-3.3.2';
const parse5SaxParserOption = {sourceCodeLocationInfo: true};
const utf8Re = /utf-?8/ui;

function ensureResponseSetHeaderWorks(res) {
	if (res === null || typeof res !== 'object' || typeof res.setHeader !== 'function') {
		const error = new TypeError(`Expected a ServerResponse object, but got ${inspectWithKind(res)}.`);

		error.code = 'ERR_INVALID_ARG_TYPE';
		Error.captureStackTrace(ensureResponseSetHeaderWorks);

		throw error;
	}

	// ref. https://nodejs.org/api/http.html#http_response_writehead_statuscode_statusmessage_headers
	// If this method is called and `response.setHeader()` has not been called,
	// it will directly write the supplied header values onto the network channel without caching internally,
	// and the `response.getHeader()` on the header will not yield the expected result.
	// If progressive population of headers is desired with potential future retrieval and modification,
	// use `response.setHeader()` instead.

	try {
		res.setHeader(TMP_HEADER_NAME, '1');
		res.removeHeader(TMP_HEADER_NAME);
	} catch {}
}

function hasHtmlContentType(res) {
	const contentTypeHeader = res.getHeader('content-type');

	try {
		return parseContentType(contentTypeHeader).type === 'text/html';
	} catch ({message}) {
		const error = new TypeError(`${contentTypeHeader} is not a valid HTTP header: ${
			message
		}.`);

		Error.captureStackTrace(error, hasHtmlContentType);
		res.emit('error', error);
	}

	return false;
}

function main(res, insertionChunk, insertionLength, etag, targetTagName, insertToEnd) { // eslint-disable-line max-params
	ensureResponseSetHeaderWorks(res);

	if (res.headersSent) {
		return;
	}

	const originalWrite = res.write.bind(res);
	const originalEnd = res.end.bind(res);
	const originalSetHeader = res.setHeader.bind(res);
	const utf8Decoder = new StringDecoder();
	const buffers = [];
	const insertionEventName = `${insertToEnd ? 'end' : 'start'}Tag`;
	let stringBuffer = '';
	let len = 0;
	let writableOffset = 0;
	let writtenOffset = 0;
	let shouldParseHtml = false;

	function onAnyToken({sourceCodeLocation: {endOffset}}) {
		writableOffset = Buffer.byteLength(stringBuffer.substring(0, endOffset));
	}

	function onTag({tagName, sourceCodeLocation: {startOffset, endOffset}}) {
		if (tagName !== targetTagName) {
			return;
		}

		for (const eventName of PARSER_EVENTS) {
			this.off(eventName, onAnyToken);
		}

		this.off(insertionEventName, onTag);
		shouldParseHtml = false;

		const insertionIndex = Buffer.byteLength(stringBuffer.substring(
			0,
			insertToEnd ? startOffset : endOffset
		)) - writtenOffset;
		let index = 0;

		for (const [arrIndex, buffer] of buffers.entries()) {
			if (index + buffer.length >= insertionIndex) {
				const sliceIndex = insertionIndex - index;

				buffers.splice(arrIndex, 1, buffer.slice(0, sliceIndex), insertionChunk, buffer.slice(sliceIndex));
				break;
			}

			index += buffer.length;
		}

		len += insertionLength;

		this.end();
	}

	const parser = new Parse5SaxParser(parse5SaxParserOption);

	for (const eventName of PARSER_EVENTS) {
		parser.on(eventName, onAnyToken);
	}

	parser.on(insertionEventName, onTag);

	function adjustContentLength(originalContentLengthHeaderValue) {
		let originalContentLength;

		if (typeof originalContentLengthHeaderValue === 'number') {
			originalContentLength = originalContentLengthHeaderValue;

			if (!Number.isInteger(originalContentLength) || originalContentLength < 0 || Object.is(originalContentLength, -0)) {
				res.emit('error', new Error(`${CONTENT_LENGTH_ERROR}, but it was ${
					inspect(originalContentLengthHeaderValue)
				}.`));

				return;
			}
		} else {
			originalContentLength = parseInt(originalContentLengthHeaderValue, 10);

			if (/\D/u.test(`${originalContentLengthHeaderValue}`)) {
				res.emit('error', new Error(`${CONTENT_LENGTH_ERROR}, but it was ${
					inspect(originalContentLengthHeaderValue)
				}.`));

				return;
			}
		}

		if (shouldParseHtml) {
			originalSetHeader('content-length', `${originalContentLength + insertionLength}`);
		}
	}

	if (res.hasHeader('content-type')) {
		shouldParseHtml = hasHtmlContentType(res);

		if (res.hasHeader('content-length')) {
			adjustContentLength(res.getHeader('content-length'));
		}
	}

	if (shouldParseHtml && res.hasHeader('etag')) {
		originalSetHeader('etag', `${res.getHeader('etag')}${etag}`);
	}

	res.setHeader = (headerName, headerValue) => {
		if (Array.isArray(headerValue)) {
			originalSetHeader(headerName, headerValue);
			return;
		}

		const lowerCaseHeaderName = headerName.toLowerCase();

		if (lowerCaseHeaderName === 'content-length') {
			adjustContentLength(headerValue);
			return;
		}

		if (shouldParseHtml && lowerCaseHeaderName === 'etag') {
			originalSetHeader(headerName, `${headerValue}${etag}`);
			return;
		}

		originalSetHeader(headerName, headerValue);

		if (lowerCaseHeaderName === 'content-type') {
			shouldParseHtml = hasHtmlContentType(res);
		}
	};

	// No need to patch `res.writeHead()` here because it calls `res.setHeader()` internally
	// https://github.com/nodejs/node/blob/v10.12.0/lib/_http_server.js#L231

	function write(data, ...restArgs) {
		const [encoding] = restArgs;

		if (shouldParseHtml && encoding && typeof encoding !== 'function' && !utf8Re.test(encoding)) {
			const error = new Error(`HTML must be UTF-8 encoded https://github.com/w3c/html/pull/1273, but encoded in ${
				encoding
			}.`);
			error.code = 'ERR_INVALID_HTML_ENCODING';

			res.emit('error', error);
			return true;
		}

		if (!Buffer.isBuffer(data)) {
			data = Buffer.from(data);
		}

		const willEnd = res.end === originalEnd;

		if (shouldParseHtml) {
			let decodedChunk;

			if (data.length !== 0) {
				len += data.length;
				buffers.push(data);
				decodedChunk = utf8Decoder.write(data);
			}

			if (willEnd) {
				if (decodedChunk) {
					stringBuffer += decodedChunk;
					parser.write(`${decodedChunk}${utf8Decoder.end()}`);
				}
			} else {
				if (decodedChunk) {
					stringBuffer += decodedChunk;
					parser.write(decodedChunk);
				}

				if (writableOffset) {
					const writableLen = writableOffset - writtenOffset;
					let index = 0;

					for (const [arrIndex, buffer] of buffers.entries()) {
						if (index + buffer.length === writableLen) {
							originalWrite(Buffer.concat(buffers.splice(0, arrIndex + 1), writableLen));
							break;
						}

						if (index + buffer.length > writableLen) {
							const sliceIndex = writableLen - index;
							const writingBuffers = buffers.splice(0, arrIndex + 1, buffer.slice(sliceIndex));

							writingBuffers.splice(-1, 1, buffer.slice(0, sliceIndex));
							originalWrite(Buffer.concat(writingBuffers, writableLen));

							break;
						}

						index += buffer.length;
					}

					writtenOffset = writableOffset;
					writableOffset = 0;
					len -= writableLen;
				}

				if (typeof restArgs[restArgs.length - 1] === 'function') {
					restArgs[restArgs.length - 1]();
				}

				return true;
			}
		} else if (willEnd) {
			if (data.length !== 0) {
				len += data.length;
				buffers.push(data);
			}
		} else {
			res.write = originalWrite;
			res.end = originalEnd;

			if (buffers.length !== 0) {
				return res.write(
					Buffer.concat([...buffers.splice(0, buffers.length), data], len + data.length),
					...restArgs
				);
			}
		}

		if (willEnd) {
			return res.end(buffers.length === 1 ? buffers[0] : Buffer.concat(buffers, len), ...restArgs);
		}

		return originalWrite(data, ...restArgs);
	}

	res.write = (...args) => {
		res.setHeader = originalSetHeader;

		if (!shouldParseHtml) {
			res.write = originalWrite;
			res.end = originalEnd;

			return res.write(...args);
		}

		if (!res.headersSent) {
			res.writeHead(res.statusCode);
		}

		res.write = write;
		return write(...args);
	};

	res.end = (data, ...restArgs) => {
		res.end = originalEnd;

		const flushed = typeof data === 'function' ? res.write(Buffer.alloc(0), data) : res.write(data || Buffer.alloc(0), ...restArgs);

		res.write = originalWrite;
		return flushed;
	};
}

function md5Base64(str) {
	return createHash('md5').update(str).digest('base64');
}

function convertInsertionChunkStringToBuffer(str) {
	if (typeof str !== 'string') {
		const error = new TypeError(`Expected a <string> to inject into HTML as the last child of \`head\` tag, but got a non-string value ${
			inspectWithKind(str)
		}.`);

		error.code = 'ERR_INVALID_ARG_TYPE';
		Error.captureStackTrace(error, convertInsertionChunkStringToBuffer);

		throw error;
	}

	return Buffer.from(str);
}

const TAG_NAME_ERROR = 'Expected `tagName` option to be an HTML tag name';
const noOptionsProvided = Symbol('noOptionsProvided');

function prepareOptionArguments(options) {
	if (options === noOptionsProvided) {
		return ['body', false];
	}

	if (!isPlainObj(options)) {
		const error = new TypeError(`Expected an <Object> to set inject-html-content options, but got ${
			inspectWithKind(options)
		}.`);
		error.code = 'ERR_INVALID_ARG_TYPE';
		Error.captureStackTrace(error, prepareOptionArguments);

		throw error;
	}

	const {tagName, insertToEnd} = options;

	if (tagName !== undefined) {
		let error;

		if (typeof tagName !== 'string') {
			error = new TypeError(`${TAG_NAME_ERROR} (<string>), but got a non-string value ${
				inspectWithKind(tagName)
			}.`);
		} else if (tagName.length === 0) {
			error = Error(`${TAG_NAME_ERROR}, but got '' (empty string).`);
		} else if (tagName.trim().length === 0) {
			error = Error(`${TAG_NAME_ERROR}, but got a whitespace-only string ${inspect(tagName)}.`);
		} else if (tagName.match(/\s/u) !== null) {
			error = Error(`${TAG_NAME_ERROR}, but got an invalid tag name ${inspect(tagName)}.`);
		}

		if (error) {
			error.code = 'ERR_INVALID_OPTION_VALUE';
			Error.captureStackTrace(error, prepareOptionArguments);

			throw error;
		}
	}

	if (insertToEnd !== undefined && typeof insertToEnd !== 'boolean') {
		const error = new TypeError(`Expected \`insertToEnd\` option to be a boolean, but got a non-boolean value ${
			inspectWithKind(insertToEnd)
		}.`);

		error.code = 'ERR_INVALID_OPTION_VALUE';
		Error.captureStackTrace(error, prepareOptionArguments);

		throw error;
	}

	return [tagName ? tagName.toLowerCase() : 'body', insertToEnd || false];
}

module.exports = function insertHtmlContent(...args) {
	const argLen = args.length;

	if (argLen !== 2 && argLen !== 3) {
		throw new RangeError(`Expected 2 or 3 arguments (<http.ServerResponse>, <string>[, <Object>]), but got ${
			argLen === 0 ? 'no' : argLen
		} arguments.`);
	}

	const [res, insertionChunk, options = noOptionsProvided] = args;
	const buffer = convertInsertionChunkStringToBuffer(insertionChunk);

	main(res, buffer, buffer.length, md5Base64(buffer), ...prepareOptionArguments(options));
};

function insertHtmlContentFromClass(...args) {
	const argLen = args.length - 5;

	if (argLen !== 1) {
		throw new RangeError(`Expected 1 argument (<http.ServerResponse>), but got ${
			argLen === 0 ? 'no' : argLen
		} arguments.`);
	}

	main(args[5], args[0], args[1], args[2], args[3], args[4]);
}

module.exports.InsertHtmlContent = class InsertHtmlContent {
	constructor(...args) {
		const argLen = args.length;

		if (argLen !== 1 && argLen !== 2) {
			throw new RangeError(`Expected 1 or 2 arguments (<string>[, <Object>]), but got ${
				argLen === 0 ? 'no' : argLen
			} arguments.`);
		}

		const [insertionChunk, options = noOptionsProvided] = args;
		const buffer = convertInsertionChunkStringToBuffer(insertionChunk);

		return insertHtmlContentFromClass.bind(
			null,
			buffer,
			buffer.length,
			md5Base64(buffer),
			...prepareOptionArguments(options)
		);
	}
};
